"""

Revision ID: 72aa496ca65b
Revises: 86457e2f333f
Create Date: 2025-06-25 17:59:28.544283

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from lecrapaud.config import LECRAPAUD_TABLE_PREFIX


# revision identifiers, used by Alembic.
revision: str = "72aa496ca65b"
down_revision: Union[str, None] = "86457e2f333f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        f"{LECRAPAUD_TABLE_PREFIX}_model_selections",
        sa.Column("best_thresholds", sa.JSON(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column(f"{LECRAPAUD_TABLE_PREFIX}_model_selections", "best_thresholds")
    # ### end Alembic commands ###
